import sys

from typing import Any, Callable, Generic, Optional, TypeVar
from twisted.python.failure import Failure

_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_F = TypeVar("_F", bound=Callable[..., Any])

def setup() -> None: ...
def run_in_reactor(
    function: Callable[..., _T]
) -> Callable[..., EventualResult[_T]]: ...

class EventualResult(Generic[_T_co]):
    def cancel(self) -> None: ...
    def wait(self, timeout: float) -> _T_co: ...
    def stash(self) -> int: ...
    def original_failure(self) -> Optional[Failure]: ...

class TimeoutError(Exception): ...

def retrieve_result(result_id: int) -> EventualResult[object]: ...
def no_setup() -> None: ...
def wait_for(timeout: float) -> Callable[[_F], _F]: ...

class ReactorStopped(Exception): ...

__version__: str



"""
Utility functions and classes.
"""

import wrapt


@wrapt.decorator
def _synced(method, self, args, kwargs):
    """Underlying synchronized wrapper."""
    with self._lock:
        return method(*args, **kwargs)


def synchronized(method):
    """
    Decorator that wraps a method with an acquire/release of self._lock.
    """
    result = _synced(method)
    result.synchronized = True
    return result



"""
Tests for crochet.mypy.
"""
from tempfile import NamedTemporaryFile
from textwrap import dedent, indent
from unittest import TestCase, skipUnless

try:
    import mypy.api
    MYPY_AVAILABLE = True
except ImportError:
    MYPY_AVAILABLE = False

MYPY_CONFIG = dedent(
    """\
    [mypy]
    plugins = crochet.mypy
    """
)


@skipUnless(MYPY_AVAILABLE, "Tests require mypy to be installed.")
class MypyTests(TestCase):
    def test_mypy_working(self) -> None:
        """
        mypy's API is able to function and produce errors when expected.
        """
        _assert_mypy(True, "ivar: int = 1\n")
        _assert_mypy(False, "ivar: int = 'bad'\n")

    def test_setup_no_args(self) -> None:
        """
        setup() and no_setup() take no arguments.
        """
        _assert_mypy(
            True,
            dedent(
                r"""
                from crochet import setup
                setup()
                """
            ),
        )
        _assert_mypy(
            True,
            dedent(
                r"""
                from crochet import no_setup
                no_setup()
                """
            ),
        )

    def test_run_in_reactor_func_takes_same_args(self) -> None:
        """
        The mypy plugin correctly passes the wrapped parameter signature through the
        @run_in_reactor decorator.
        """
        template = dedent(
            """\
            from crochet import run_in_reactor

            @run_in_reactor
            def foo({params}) -> None:
                pass

            foo({args})
            """
        )
        for params, args, good in (
            (
                "x: int, y: str, z: float, *a: int, **kw: str",
                "1, 'something', -1, 4, 5, 6, k1='x', k2='y'",
                True,
            ),
            (
                "",
                "1",
                False,
            ),
            (
                "x: int",
                "",
                False,
            ),
            (
                "x: int",
                "1, 2",
                False,
            ),
            (
                "x: int",
                "'something'",
                False,
            ),
            (
                "*x: int",
                "1, 2, 3",
                True,
            ),
            (
                "*x: int",
                "'something'",
                False,
            ),
            (
                "**x: int",
                "k1=16, k2=-5",
                True,
            ),
            (
                "**x: int",
                "k1='something'",
                False,
            ),
            (
                "x: int, y: str",
                "1, 'ok'",
                True,
            ),
            (
                "x: int, y: str",
                "'not ok', 1",
                False,
            ),
            (
                "x: str, y: int",
                "'ok', 1",
                True,
            ),
            (
                "x: str, y: int",
                "1, 'not ok'",
                False,
            ),
        ):
            with self.subTest(params=params, args=args):
                _assert_mypy(good, template.format(params=params, args=args))

    def test_run_in_reactor_func_returns_typed_eventual(self) -> None:
        """
        run_in_reactor preserves the decorated function's return type indirectly
        through an EventualResult.
        """
        template = dedent(
            """\
            from typing import Optional
            from crochet import EventualResult, run_in_reactor

            @run_in_reactor
            def foo() -> {return_type}:
                return {return_value}

            eventual_result: {receiver_type} = foo()
            final_result: {final_type} = eventual_result.wait(1)
            """
        )
        for return_type, return_value, receiver_type, final_type, good in (
            (
                "int",
                "1",
                "EventualResult[int]",
                "int",
                True,
            ),
            (
                "int",
                "'str'",
                "EventualResult[int]",
                "int",
                False,
            ),
            (
                "int",
                "1",
                "EventualResult[str]",
                "int",
                False,
            ),
            (
                "int",
                "1",
                "EventualResult[str]",
                "str",
                False,
            ),
            (
                "int",
                "1",
                "int",
                "int",
                False,
            ),
            (
                "int",
                "1",
                "EventualResult[int]",
                "Optional[int]",
                True,
            ),
            (
                "Optional[int]",
                "1",
                "EventualResult[Optional[int]]",
                "Optional[int]",
                True,
            ),
            (
                "Optional[int]",
                "None",
                "EventualResult[Optional[int]]",
                "Optional[int]",
                True,
            ),
            (
                "Optional[int]",
                "1",
                "EventualResult[int]",
                "Optional[int]",
                False,
            ),
            (
                "Optional[int]",
                "1",
                "EventualResult[Optional[int]]",
                "int",
                False,
            ),
        ):
            with self.subTest(
                return_type=return_type,
                return_value=return_value,
                receiver_type=receiver_type,
                final_type=final_type,
            ):
                _assert_mypy(
                    good,
                    template.format(
                        return_type=return_type,
                        return_value=return_value,
                        receiver_type=receiver_type,
                        final_type=final_type,
                    ),
                )

    def test_run_in_reactor_func_signature_transform(self) -> None:
        """
        The mypy plugin correctly passes the wrapped signature though the
        @run_in_reactor decorator with an EventualResult-wrapped return type.
        """
        template = dedent(
            """\
            from typing import Callable
            from crochet import EventualResult, run_in_reactor

            class Thing:
                pass

            @run_in_reactor
            def foo(x: int, y: str, z: float) -> Thing:
                return Thing()

            re_foo: {result_type} = foo
            """
        )
        for result_type, good in (
            ("Callable[[int, str, float], EventualResult[Thing]]", True),
            ("Callable[[int, str, float], EventualResult[object]]", True),
            ("Callable[[int, str, float], EventualResult[int]]", False),
            ("Callable[[int, str, float], Thing]", False),
            ("Callable[[int, str, float], int]", False),
            ("Callable[[int, str], EventualResult[Thing]]", False),
            ("Callable[[int], EventualResult[Thing]]", False),
            ("Callable[[], EventualResult[Thing]]", False),
            ("Callable[[float, int, str], EventualResult[Thing]]", False),
        ):
            with self.subTest(result_type=result_type):
                _assert_mypy(good, template.format(result_type=result_type))

    def test_eventual_result_cancel_signature(self) -> None:
        """
        EventualResult's cancel() method takes no arguments.
        """
        _assert_mypy(
            True,
            dedent(
                """\
                from crochet import EventualResult
                def foo(er: EventualResult[object]) -> None:
                    er.cancel()
                """
            ),
        )

    def test_eventual_result_wait_signature(self) -> None:
        """
        EventualResult's wait() method takes one timeout float argument.
        """
        _assert_mypy(
            True,
            dedent(
                """\
                from crochet import EventualResult
                def foo(er: EventualResult[object]) -> object:
                    return er.wait(2.0)
                """
            ),
        )
        _assert_mypy(
            True,
            dedent(
                """\
                from crochet import EventualResult
                def foo(er: EventualResult[object]) -> object:
                    return er.wait(timeout=2.0)
                """
            ),
        )

    def test_eventual_result_stash_signature(self) -> None:
        """
        EventualResult's stash() method takes no arguments and returns the same type
        retrieve_result's one result_id parameter takes.
        """
        _assert_mypy(
            True,
            dedent(
                """\
                from crochet import EventualResult, retrieve_result
                def foo(er: EventualResult[object]) -> None:
                    retrieve_result(er.stash())
                    retrieve_result(result_id=er.stash())
                """
            ),
        )

    def test_eventual_result_original_failure_signature(self) -> None:
        """
        EventualResult's original_failure() method takes no arguments and returns an
        optional Failure.
        """
        _assert_mypy(
            True,
            dedent(
                """\
                from typing import Optional
                from twisted.python.failure import Failure
                from crochet import EventualResult
                def foo(er: EventualResult[object]) -> Optional[Failure]:
                    return er.original_failure()
                """
            ),
        )
        _assert_mypy(
            False,
            dedent(
                """\
                from twisted.python.failure import Failure
                from crochet import EventualResult
                def foo(er: EventualResult[object]) -> Failure:
                    return er.original_failure()
                """
            ),
        )

    def test_exceptions(self) -> None:
        """
        ReactorStopped and TimeoutError are Exception types.
        """
        _assert_mypy(
            True,
            dedent(
                """\
                from crochet import ReactorStopped, TimeoutError
                e1: Exception = ReactorStopped()
                e2: Exception = TimeoutError()
                """
            ),
        )

    def test_retrieve_result_returns_untyped_eventual_result(self) -> None:
        """
        retrieve_result() returns an untyped EventualResult.
        """
        _assert_mypy(
            True,
            dedent(
                """\
                from crochet import EventualResult, retrieve_result
                r: EventualResult[object] = retrieve_result(3)
                """
            ),
        )
        _assert_mypy(
            False,
            dedent(
                """\
                from crochet import EventualResult, retrieve_result
                r: EventualResult[int] = retrieve_result(3)
                """
            ),
        )

    def test_wait_for_signature(self) -> None:
        """
        The @wait_for decorator takes a timeout float.
        """
        _assert_mypy(
            True,
            dedent(
                """\
                from crochet import wait_for

                @wait_for(1.5)
                def foo() -> None:
                    pass
                """
            ),
        )
        _assert_mypy(
            True,
            dedent(
                """\
                from crochet import wait_for

                @wait_for(timeout=1.5)
                def foo() -> None:
                    pass
                """
            ),
        )

    def test_wait_for_func_signature_unchanged(self) -> None:
        """
        The @wait_for(timeout) decorator preserves the wrapped function's signature.
        """
        template = dedent(
            """\
            from typing import Callable
            from crochet import wait_for

            class Thing:
                pass

            @wait_for(1)
            def foo(x: int, y: str, z: float) -> Thing:
                return Thing()

            re_foo: {result_type} = foo
            """
        )
        for result_type, good in (
            ("Callable[[int, str, float], Thing]", True),
            ("Callable[[int, str, float], object]", True),
            ("Callable[[int, str, float], int]", False),
            ("Callable[[int, str, float], EventualResult[Thing]]", False),
            ("Callable[[int, str, float], None]", False),
            ("Callable[[int, str], Thing]", False),
            ("Callable[[int], Thing]", False),
            ("Callable[[], Thing]", False),
            ("Callable[[float, int, str], Thing]", False),
        ):
            with self.subTest(result_type=result_type):
                _assert_mypy(good, template.format(result_type=result_type))

    def test_version_string(self) -> None:
        """
        __version__ is a string.
        """
        _assert_mypy(
            True,
            dedent(
                """\
                import crochet
                x: str = crochet.__version__
                """
            ),
        )


def _assert_mypy(expect_success: bool, source_code: str) -> None:
    with NamedTemporaryFile(mode="w+t", delete=False) as config_file:
        config_file.write(MYPY_CONFIG)

    out, err, status = mypy.api.run(
        ["--config-file", config_file.name, "-c", source_code]
    )
    if status not in (0, 1):
        raise RuntimeError(
            f"Unexpected mypy error (status {status}):\n{indent(err, ' ' * 2)}"
        )
    if expect_success:
        assert (
            status == 0
        ), f"Unexpected mypy failure (status {status}):\n{indent(out, ' ' * 2)}"
    else:
        assert status == 1, f"Unexpected mypy success: stdout: {out}\nstderr: {err}\n"



"""
Mypy plugin to aid with typechecking code that uses Crochet.
"""

import typing
from typing import Callable, Optional

from mypy.plugin import FunctionContext, Plugin  # pylint: disable=no-name-in-module
from mypy.types import CallableType, Type, get_proper_type  # pylint: disable=no-name-in-module


def plugin(_version: str) -> typing.Type[Plugin]:
    return CrochetMypyPlugin


class CrochetMypyPlugin(Plugin):
    """
    Assists mypy with type checking APIs not (yet) fully covered by Python's
    type hint annotation types, by copying run_in_reactor decorated function's
    argument types to the type mypy deduces for the wrapped function.
    """

    def get_function_hook(
        self,
        fullname: str,
    ) -> Optional[Callable[[FunctionContext], Type]]:
        if fullname == "crochet.run_in_reactor":
            return _copyargs_callback
        return None


def _copyargs_callback(ctx: FunctionContext) -> Type:
    """
    Copy the parameters from the signature of the type of the argument of the
    call to the signature of the return type.
    """
    original_return_type = ctx.default_return_type
    if not ctx.arg_types or len(ctx.arg_types[0]) != 1:
        return original_return_type

    arg_type = get_proper_type(ctx.arg_types[0][0])
    default_return_type = get_proper_type(original_return_type)

    if not (
        isinstance(arg_type, CallableType)
        and isinstance(default_return_type, CallableType)
    ):
        return original_return_type

    return default_return_type.copy_modified(
        arg_types=arg_type.arg_types,
        arg_kinds=arg_type.arg_kinds,
        arg_names=arg_type.arg_names,
        variables=arg_type.variables,
        is_ellipsis_args=arg_type.is_ellipsis_args,
    )



"""
Tests for the initial setup.
"""

from __future__ import absolute_import

import threading
import warnings
import subprocess
import sys
from unittest import SkipTest, TestCase

import twisted
from twisted.python.log import PythonLoggingObserver
from twisted.python import log
from twisted.python.runtime import platform
from twisted.internet.task import Clock

from .._eventloop import EventLoop, ThreadLogObserver, _store
from ..tests import crochet_directory


class FakeReactor(Clock):
    """
    A fake reactor for testing purposes.
    """
    thread_id = None
    runs = 0
    in_call_from_thread = False

    def __init__(self):
        Clock.__init__(self)
        self.started = threading.Event()
        self.stopping = False
        self.events = []

    def run(self, installSignalHandlers=True):
        self.runs += 1
        self.thread_id = threading.current_thread().ident
        self.installSignalHandlers = installSignalHandlers
        self.started.set()

    def callFromThread(self, f, *args, **kwargs):
        self.in_call_from_thread = True
        f(*args, **kwargs)
        self.in_call_from_thread = False

    def stop(self):
        self.stopping = True

    def addSystemEventTrigger(self, when, event, f):
        self.events.append((when, event, f))


class FakeThread:
    started = False

    def start(self):
        self.started = True


class SetupTests(TestCase):
    """
    Tests for setup().
    """

    def test_first_runs_reactor(self):
        """
        With it first call, setup() runs the reactor in a thread.
        """
        reactor = FakeReactor()
        EventLoop(lambda: reactor, lambda f, *g: None).setup()
        reactor.started.wait(5)
        self.assertNotEqual(reactor.thread_id, None)
        self.assertNotEqual(
            reactor.thread_id, threading.current_thread().ident)
        self.assertFalse(reactor.installSignalHandlers)

    def test_second_does_nothing(self):
        """
        The second call to setup() does nothing.
        """
        reactor = FakeReactor()
        s = EventLoop(lambda: reactor, lambda f, *g: None)
        s.setup()
        s.setup()
        reactor.started.wait(5)
        self.assertEqual(reactor.runs, 1)

    def test_stop_on_exit(self):
        """
        setup() registers an exit handler that stops the reactor, and an exit
        handler that logs stashed EventualResults.
        """
        atexit = []
        reactor = FakeReactor()
        s = EventLoop(
            lambda: reactor, lambda f, *args: atexit.append((f, args)))
        s.setup()
        self.assertEqual(len(atexit), 2)
        self.assertFalse(reactor.stopping)
        f, args = atexit[0]
        self.assertEqual(f, reactor.callFromThread)
        self.assertEqual(args, (reactor.stop, ))
        f(*args)
        self.assertTrue(reactor.stopping)
        f, args = atexit[1]
        self.assertEqual(f, _store.log_errors)
        self.assertEqual(args, ())
        f(*args)  # make sure it doesn't throw an exception

    def test_runs_with_lock(self):
        """
        All code in setup() and no_setup() is protected by a lock.
        """
        self.assertTrue(EventLoop.setup.synchronized)
        self.assertTrue(EventLoop.no_setup.synchronized)

    def test_logging(self):
        """
        setup() registers a PythonLoggingObserver wrapped in a
        ThreadLogObserver, removing the default log observer.
        """
        logging = []

        def fakeStartLoggingWithObserver(observer, setStdout=1):
            self.assertIsInstance(observer, ThreadLogObserver)
            wrapped = observer._observer
            expected = PythonLoggingObserver.emit
            # Python 3 and 2 differ in value of __func__:
            expected = getattr(expected, "__func__", expected)
            self.assertIs(wrapped.__func__, expected)
            self.assertEqual(setStdout, False)
            self.assertTrue(reactor.in_call_from_thread)
            logging.append(observer)

        reactor = FakeReactor()
        loop = EventLoop(
            lambda: reactor, lambda f, *g: None, fakeStartLoggingWithObserver)
        loop.setup()
        self.assertTrue(logging)
        logging[0].stop()

    def test_stop_logging_on_exit(self):
        """
        setup() registers a reactor shutdown event that stops the logging
        thread.
        """
        observers = []
        reactor = FakeReactor()
        s = EventLoop(
            lambda: reactor,
            lambda f, *arg: None,
            lambda observer, setStdout=1: observers.append(observer))
        s.setup()
        self.addCleanup(observers[0].stop)
        self.assertIn(("after", "shutdown", observers[0].stop), reactor.events)

    def test_warnings_untouched(self):
        """
        setup() ensure the warnings module's showwarning is unmodified,
        overriding the change made by normal Twisted logging setup.
        """

        def fakeStartLoggingWithObserver(observer, setStdout=1):
            warnings.showwarning = log.showwarning
            self.addCleanup(observer.stop)

        original = warnings.showwarning
        reactor = FakeReactor()
        loop = EventLoop(
            lambda: reactor, lambda f, *g: None, fakeStartLoggingWithObserver)
        loop.setup()
        self.assertIs(warnings.showwarning, original)

    def test_start_watchdog_thread(self):
        """
        setup() starts the shutdown watchdog thread.
        """
        thread = FakeThread()
        reactor = FakeReactor()
        loop = EventLoop(
            lambda: reactor, lambda *args: None, watchdog_thread=thread)
        loop.setup()
        self.assertTrue(thread.started)

    def test_no_setup(self):
        """
        If called first, no_setup() makes subsequent calls to setup() do
        nothing.
        """
        observers = []
        atexit = []
        thread = FakeThread()
        reactor = FakeReactor()
        loop = EventLoop(
            lambda: reactor,
            lambda f, *arg: atexit.append(f),
            lambda observer, *a, **kw: observers.append(observer),
            watchdog_thread=thread)

        loop.no_setup()
        loop.setup()
        self.assertFalse(observers)
        self.assertFalse(atexit)
        self.assertFalse(reactor.runs)
        self.assertFalse(thread.started)

    def test_no_setup_after_setup(self):
        """
        If called after setup(), no_setup() throws an exception.
        """
        reactor = FakeReactor()
        s = EventLoop(lambda: reactor, lambda f, *g: None)
        s.setup()
        self.assertRaises(RuntimeError, s.no_setup)

    def test_setup_registry_shutdown(self):
        """
        ResultRegistry.stop() is registered to run before reactor shutdown by
        setup().
        """
        reactor = FakeReactor()
        s = EventLoop(lambda: reactor, lambda f, *g: None)
        s.setup()
        self.assertEqual(
            reactor.events, [("before", "shutdown", s._registry.stop)])

    def test_no_setup_registry_shutdown(self):
        """
        ResultRegistry.stop() is registered to run before reactor shutdown by
        setup().
        """
        reactor = FakeReactor()
        s = EventLoop(lambda: reactor, lambda f, *g: None)
        s.no_setup()
        self.assertEqual(
            reactor.events, [("before", "shutdown", s._registry.stop)])


class ProcessSetupTests(TestCase):
    """
    setup() enables support for IReactorProcess on POSIX plaforms.
    """

    def test_posix(self):
        """
        On POSIX systems, setup() installs a LoopingCall that runs
        t.i.process.reapAllProcesses() 10 times a second.
        """
        if platform.type != "posix":
            raise SkipTest("SIGCHLD is a POSIX-specific issue")
        reactor = FakeReactor()
        reaps = []
        s = EventLoop(
            lambda: reactor,
            lambda f, *g: None,
            reapAllProcesses=lambda: reaps.append(1))
        s.setup()
        reactor.advance(0.1)
        self.assertEqual(reaps, [1])
        reactor.advance(0.1)
        self.assertEqual(reaps, [1, 1])
        reactor.advance(0.1)
        self.assertEqual(reaps, [1, 1, 1])

    def test_non_posix(self):
        """
        On non-POSIX systems, setup() does not install a LoopingCall.
        """
        if platform.type == "posix":
            raise SkipTest("This test is for non-POSIX systems.")
        reactor = FakeReactor()
        s = EventLoop(lambda: reactor, lambda f, *g: None)
        s.setup()
        self.assertFalse(reactor.getDelayedCalls())


class ReactorImportTests(TestCase):
    """
    Tests for when the reactor gets imported.

    The reactor should only be imported as part of setup()/no_setup(),
    rather than as side-effect of Crochet import, since daemonization
    doesn't work if reactor is imported
    (https://twistedmatrix.com/trac/ticket/7105).
    """

    def test_crochet_import_no_reactor(self):
        """
        Importing crochet should not import the reactor.
        """
        program = """\
import sys
import crochet

if "twisted.internet.reactor" not in sys.modules:
    sys.exit(23)
"""
        process = subprocess.Popen([sys.executable, "-c", program],
                                   cwd=crochet_directory)
        self.assertEqual(process.wait(), 23)


LOGGING_PROGRAM = """\
import sys
from logging import StreamHandler, Formatter, getLogger, DEBUG
handler = StreamHandler(sys.stdout)
handler.setFormatter(Formatter("%%(levelname)s %%(message)s"))
l = getLogger("twisted")
l.addHandler(handler)
l.setLevel(DEBUG)

import crochet
crochet.setup()
from twisted.python import log
%s
log.msg("log-info")
log.msg("log-error", isError=True)
"""


class LoggingTests(TestCase):
    """
    End-to-end tests for Twisted->stdlib logging bridge.
    """
    maxDiff = None

    def test_old_logging(self):
        """
        Messages from the old Twisted logging API are emitted to Python
        standard library logging.
        """
        if tuple(map(int, twisted.__version__.split("."))) >= (15, 2, 0):
            raise SkipTest("This test is for Twisted < 15.2.")

        program = LOGGING_PROGRAM % ("", )
        output = subprocess.check_output([sys.executable, "-u", "-c", program],
                                         cwd=crochet_directory)
        self.assertTrue(
            output.startswith(
                """\
INFO Log opened.
INFO log-info
ERROR log-error
"""))

    def test_new_logging(self):
        """
        Messages from both new and old Twisted logging APIs are emitted to
        Python standard library logging.
        """
        if tuple(map(int, twisted.__version__.split("."))) < (15, 2, 0):
            raise SkipTest("This test is for Twisted 15.2 and later.")

        program = LOGGING_PROGRAM % (
            """\
from twisted.logger import Logger
l2 = Logger()
import time
time.sleep(1)  # workaround, there is race condition... somewhere
l2.info("logger-info")
l2.critical("logger-critical")
l2.warn("logger-warning")
l2.debug("logger-debug")
""", )
        output = subprocess.check_output([sys.executable, "-u", "-c", program],
                                         cwd=crochet_directory)
        self.assertIn(
            """\
INFO logger-info
CRITICAL logger-critical
WARNING logger-warning
DEBUG logger-debug
INFO log-info
CRITICAL log-error
""", output.decode("utf-8").replace("\r\n", "\n"))
